import pandas as pd
from polarbear import entities
from savReaderWriter import SavReader, SavHeaderReader
from sqlalchemy import and_, select, MetaData


def read_db(engine, study_ids):

    question_ids = get_question_ids_based_on_study_ids_from_db(engine, study_ids)
    question_metadata = read_question_metadata_from_db(
        engine, question_ids)
    codebook = read_codebook_from_db(engine, question_ids)
    data = read_data_from_db(engine, study_ids, question_ids,
                             question_metadata)

    ds = entities.Dataset(data, codebook, question_metadata)
    return ds


def get_question_ids_based_on_study_ids_from_db(engine, study_ids):

    conn = engine.connect()
    meta = MetaData(engine, reflect=True)

    question_orders = meta.tables['question_orders']
    s = select([question_orders.c.question_id]).where(
        question_orders.c.study_id.in_(study_ids))
    question_ids = list(pd.read_sql(s, conn).question_id)

    return question_ids


def read_question_metadata_from_db(engine, question_ids):
    conn = engine.connect()
    meta = MetaData(engine, reflect=True)

    questions = meta.tables['questions']
    s = select([questions]).where(questions.c.question_id.in_(question_ids))
    question_metadata = pd.read_sql(s, conn)
    scale_ids = list(question_metadata.scale_id.unique())

    scales = meta.tables['scales']
    s = select([scales]).where(scales.c.scale_id.in_(scale_ids))
    scale_metadata = pd.read_sql(s, conn)

    question_metadata = question_metadata.merge(
        scale_metadata, how='left', on=['scale_id'])

    return question_metadata


def read_codebook_from_db(engine, question_ids):
    conn = engine.connect()
    meta = MetaData(engine, reflect=True)

    response_choices = meta.tables['response_choices']
    s = select([response_choices]).where(
        response_choices.c.question_id.in_(question_ids))

    codebook = pd.read_sql(s, conn)[[
        'question_id', 'full_label', 'display_label', 'value']]

    codebook.rename(index=str, columns={'value': 'value_code'}, inplace=True)

    return codebook


def read_data_from_db(engine, study_ids, question_ids,
                      question_metadata=None):
    conn = engine.connect()
    meta = MetaData(engine, reflect=True)

    closed_ended_responses = meta.tables['closed_ended_responses']
    s = select([closed_ended_responses]).where((
        and_(closed_ended_responses.c.question_id.in_(question_ids),
             closed_ended_responses.c.study_id.in_(study_ids))))
    closed_ended_data = pd.read_sql(s, conn)
    closed_ended_data = closed_ended_data.pivot(
        index='respondent_id', columns='question_id', values='value')

    open_ended_responses = meta.tables['open_ended_responses']
    s = select([open_ended_responses]).where((
        and_(open_ended_responses.c.question_id.in_(question_ids),
             open_ended_responses.c.study_id.in_(study_ids))))
    open_ended_data = pd.read_sql(s, conn)
    open_ended_data = open_ended_data.pivot(
        index='respondent_id', columns='question_id', values='text_response')

    data = closed_ended_data.merge(
        open_ended_data, how='left', left_index=True, right_index=True)

    data.reset_index(inplace=True)

    if question_metadata is None:
        question_metadata = read_question_metadata_from_db(
            engine, question_ids)

    for i, row in question_metadata.iterrows():
        if row['type_of_response'] == 'Date':
            data[row['question_id']] = pd.to_datetime(data[row['question_id']])

    return data


def _handle_read_spss_datetime(data, datetime_cols=None,
                               date_cols=None, time_cols=None):

    if datetime_cols is not None:
        for col in datetime_cols:
            data[col] = pd.to_datetime(
                data[col]-7456579200, unit='s',
                origin=pd.to_datetime('1819-01-28'))

    if date_cols is not None:
        for col in date_cols:
            data[col] = pd.to_datetime(
                data[col]-7456579200, unit='s',
                origin=pd.to_datetime('1819-01-28')).dt.date

    if time_cols is not None:
        for col in time_cols:
            data[col] = pd.to_datetime(data[col], unit='s').dt.time

    return data


def read_spss(filepath, ioUtf8=True, datetime_cols=None,
              date_cols=None, time_cols=None, *args, **kwargs):

    with SavReader(filepath, ioUtf8=ioUtf8, *args, **kwargs) as reader:
        cols = reader.header
        d = reader.all()
    data = pd.DataFrame(columns=cols, data=d).replace(
        ['\r', '\n'], '', regex=True)

    if not ((datetime_cols is None) & (date_cols is None) &
            (time_cols is None)):
        data = _handle_read_spss_datetime(data,
                                          datetime_cols=datetime_cols,
                                          date_cols=date_cols,
                                          time_cols=time_cols)

    with SavHeaderReader(filepath, ioUtf8=ioUtf8) as header:
        var_names = [v for v in header.all().varNames]
        var_labels = ({k: v for
                       k, v in header.all().varLabels.items()})
        value_labels = header.all().valueLabels
        value_labels = [[[question,
                          value,
                          label]
                        for value, label in value_label_map.items()] for
                        question, value_label_map in value_labels.items()]
        value_labels = [value_label_map for question in value_labels
                        for value_label_map in question]
        codebook = pd.DataFrame(columns=['question_id', 'value_code',
                                         'full_label'],
                                data=value_labels)
        codebook['display_label'] = codebook.full_label
        m = pd.DataFrame(
            columns=['question_id', 'full_text', 'display_text',
                     'category', 'subcategory', 'topic', 'subtopic',
                     'type_of_response',
                     'scale_id', 'scale_name', 'type_of_scale']
        )
        m.question_id = var_names
        m.full_text = m.question_id.replace(var_labels)
        question_metadata = m.replace(['\r', '\n'], '', regex=True)

    ds = entities.Dataset(data, codebook, question_metadata)
    return ds


def read_csv(filepath, codebook_path=None,
             question_metadata_path=None, *args, **kwargs):
    data = pd.read_csv(filepath, *args, **kwargs)

    if codebook_path is not None:
        codebook = pd.read_csv(codebook_path, *args, **kwargs)
    else:
        codebook = None

    if question_metadata_path is not None:
        question_metadata = pd.read_csv(
            question_metadata_path, *args, **kwargs)
    else:
        question_metadata = None

    ds = entities.Dataset(data, codebook, question_metadata)
    return ds


def read_excel(filepath, codebook_path=None,
               question_metadata_path=None, *args, **kwargs):
    data = pd.read_excel(filepath, *args, **kwargs)
    for col in data.columns:
        if data[col].dtypes == int:
            data[col] = data[col].astype(float)

    data.apply(lambda x: x.astype(float) if x.dtype == str else x,
               axis=1)

    if codebook_path is not None:
        codebook = pd.read_excel(codebook_path, dtype={
            'question_id': str,
            'value_code': float,
            'full_label': str,
            'display_label': str
        }, *args, **kwargs)
    else:
        codebook = None

    if question_metadata_path is not None:
        question_metadata = pd.read_excel(
            question_metadata_path, *args, **kwargs)
    else:
        question_metadata = None

    ds = entities.Dataset(data, codebook, question_metadata)
    return ds


def read_schema_from_csv(schema_studies_filepath,
                         schema_question_orders_filepath,
                         schema_questions_filepath,
                         schema_scales_filepath,
                         schema_response_choices_filepath,
                         schema_closed_ended_responses_filepath,
                         schema_open_ended_responses_filepath
                         ):
    studies = pd.read_csv(schema_studies_filepath)
    studies['start_date'] = pd.to_datetime(studies['start_date'])
    studies['end_date'] = pd.to_datetime(studies['end_date'])

    question_orders = pd.read_csv(schema_question_orders_filepath)
    question_orders['order'] = question_orders['order'].astype(float)

    questions = pd.read_csv(schema_questions_filepath)
    scales = pd.read_csv(schema_scales_filepath)

    response_choices = pd.read_csv(schema_response_choices_filepath)
    response_choices['value'] = response_choices['value'].astype(float)

    closed_ended_responses = pd.read_csv(
        schema_closed_ended_responses_filepath)
    closed_ended_responses['value'] = closed_ended_responses['value'].astype(float)

    open_ended_responses = pd.read_csv(schema_open_ended_responses_filepath)

    schema = entities.Schema(
        studies, question_orders, questions, scales,
        response_choices, closed_ended_responses, open_ended_responses)

    return schema


def read_schema_from_excel(schema_studies_filepath,
                           schema_question_orders_filepath,
                           schema_questions_filepath,
                           schema_scales_filepath,
                           schema_response_choices_filepath,
                           schema_closed_ended_responses_filepath,
                           schema_open_ended_responses_filepath
                           ):
    studies = pd.read_excel(schema_studies_filepath)
    studies['start_date'] = pd.to_datetime(studies['start_date'])
    studies['end_date'] = pd.to_datetime(studies['end_date'])

    question_orders = pd.read_excel(schema_question_orders_filepath)
    question_orders['order'] = question_orders['order'].astype(float)

    questions = pd.read_excel(schema_questions_filepath)
    scales = pd.read_excel(schema_scales_filepath)

    response_choices = pd.read_excel(schema_response_choices_filepath)
    response_choices['value'] = response_choices['value'].astype(float)

    closed_ended_responses = pd.read_excel(
        schema_closed_ended_responses_filepath)
    closed_ended_responses['value'] = closed_ended_responses['value'].astype(float)

    open_ended_responses = pd.read_excel(schema_open_ended_responses_filepath)

    schema = entities.Schema(
        studies, question_orders, questions, scales,
        response_choices, closed_ended_responses, open_ended_responses)

    return schema



import pandas as pd
from savReaderWriter import SavWriter


class Mapping(object):

    def __init__(self, study_id, info=None):
        self.study_id = study_id

        if info is None:
            self.info = pd.DataFrame(
                columns=['question_id', 'database_question_id',
                         'exists_in_database', 'full_text', 'type_of_response',
                         'type_of_scale', 'scale_exists_in_database',
                         'scale_id', 'scale_name', 'display_text',
                         'category', 'subcategory', 'topic', 'subtopic'],
                data=[['id_in_study', None, None,
                       None, None, None, None, None, None,
                       None, None, None, None, None]])

        else:
            assert isinstance(info, pd.DataFrame), (
                "info argument should be a pandas DataFrame")
            assert "question_id" in info.columns, (
                "info should contain question_id column")
            assert "database_question_id" in info.columns, (
                "info should contain database_question_id column")
            assert "exists_in_database" in info.columns, (
                "info should contain exists_in_database column")
            assert "full_text" in info.columns, (
                "info should contain full_text column")
            assert "type_of_response" in info.columns, (
                "info should contain type_of_response column")
            assert "type_of_scale" in info.columns, (
                "info should contain type_of_scale column")
            assert "scale_exists_in_database" in info.columns, (
                "info should contain scale_exists_in_database column")
            assert "scale_id" in info.columns, (
                "info should contain scale_id column")
            assert "scale_name" in info.columns, (
                "info should contain scale_name column")
            assert "display_text" in info.columns, (
                "info should contain display_text column")
            assert "category" in info.columns, (
                "info should contain category column")
            assert "subcategory" in info.columns, (
                "info should contain subcategory column")
            assert "topic" in info.columns, (
                "info should contain topic column")
            assert "subtopic" in info.columns, (
                "info should contain subtopic column")
            assert "id_in_study" in list(info.question_id), (
                "question_id column in mapping should contain row for id_in_study")
            self.info = info

    def check_schema_conversion_readiness(self):
        assert isinstance(self.info, pd.DataFrame), (
            "info should be a pandas DataFrame")
        assert "question_id" in self.info.columns, (
            "info should contain question_id column")
        assert "database_question_id" in self.info.columns, (
            "info should contain database_question_id column")
        assert "exists_in_database" in self.info.columns, (
            "info should contain exists_in_database column")
        assert "full_text" in self.info.columns, (
            "info should contain full_text column")
        assert "type_of_response" in self.info.columns, (
            "info should contain type_of_response column")
        assert "type_of_scale" in self.info.columns, (
            "info should contain type_of_scale column")
        assert "scale_exists_in_database" in self.info.columns, (
            "info should contain scale_exists_in_database column")
        assert "scale_id" in self.info.columns, (
            "info should contain scale_id column")
        assert "scale_name" in self.info.columns, (
            "info should contain scale_name column")
        assert "display_text" in self.info.columns, (
            "info should contain display_text column")
        assert "category" in self.info.columns, (
            "info should contain category column")
        assert "subcategory" in self.info.columns, (
            "info should contain subcategory column")
        assert "topic" in self.info.columns, (
            "info should contain topic column")
        assert "subtopic" in self.info.columns, (
            "info should contain subtopic column")


class Dataset(object):

    def __init__(self, data, codebook=None, question_metadata=None):

        assert isinstance(data, pd.DataFrame), (
            "data argument should be a pandas DataFrame")
        self.data = data.apply(
            lambda x: x.astype(float, errors='ignore'), axis=1)

        if codebook is None:
            self.codebook = pd.DataFrame(columns=['question_id', 'value_code',
                                         'full_label', 'display_label'],
                                         dtype=float)
        else:
            assert isinstance(codebook, pd.DataFrame), (
                "codebook argument should be a pandas DataFrame")
            assert "question_id" in codebook.columns, (
                "codebook should contain question_id column")
            assert "value_code" in codebook.columns, (
                "codebook should contain value_code column")
            assert "full_label" in codebook.columns, (
                "codebook should contain full_label column")
            self.codebook = codebook
            self.codebook.value_code = self.codebook.value_code.astype(float)

        if question_metadata is None:
            m = pd.DataFrame(
                columns=['question_id', 'full_text', 'display_text',
                         'category', 'subcategory', 'topic', 'subtopic',
                         'type_of_response',
                         'scale_id', 'scale_name', 'type_of_scale'],
                dtype=float
            )
            m.question_id = data.columns
            self.question_metadata = m
        else:
            assert isinstance(question_metadata, pd.DataFrame), (
                "question_metadata argument should be a pandas DataFrame")
            assert "question_id" in question_metadata.columns, (
                "question_metadata should contain question_id column")
            self.question_metadata = question_metadata

    def add_id_in_study(self):
        self.data.insert(loc=0, column="id_in_study",
                         value=range(1, len(self.data) + 1))
        return self

    def add_value_code_to_codebook(self, questions, value,
                                   full_label, display_label):

        df = pd.DataFrame({'question_id': questions,
                           'value_code': value,
                           'full_label': full_label,
                           'display_label': display_label})

        self.codebook = self.codebook.append(df, ignore_index=True)

        return self

    def to_csv(self, filepath, codebook_path=None,
               question_metadata_path=None, encoding='utf-8', *args, **kwargs):
        self.data.to_csv(filepath, index=False, encoding=encoding,
                         *args, **kwargs)

        if codebook_path is None:
            codebook_path = filepath.replace('.csv', '_codebook.csv')
        else:
            codebook_path = codebook_path
        self.codebook.to_csv(codebook_path, index=False, encoding=encoding,
                             *args, **kwargs)

        if question_metadata_path is None:
            question_metadata_path = filepath.replace(
                '.csv', '_question_metadata.csv')
        else:
            question_metadata_path = question_metadata_path
        self.question_metadata.to_csv(question_metadata_path, index=False,
                                      encoding=encoding, *args, **kwargs)

    def to_excel(self, filepath, codebook_path=None,
                 question_metadata_path=None, encoding='utf-8',
                 *args, **kwargs):
        self.data.to_excel(filepath, index=False, encoding=encoding,
                           *args, **kwargs)

        if codebook_path is None:
            codebook_path = filepath.replace('.xlsx', '_codebook.xlsx')
        else:
            codebook_path = codebook_path
        self.codebook.to_excel(codebook_path, index=False, encoding=encoding,
                               *args, **kwargs)

        if question_metadata_path is None:
            question_metadata_path = filepath.replace(
                '.xlsx', '_question_metadata.xlsx')
        else:
            question_metadata_path = question_metadata_path
        self.question_metadata.to_excel(question_metadata_path, index=False,
                                        encoding=encoding, *args, **kwargs)

    def to_spss(self, filepath, *args, **kwargs):
        columns = self.question_metadata.question_id
        var_names = list(columns)
        var_types = {
            c: (0 if pd.api.types.is_numeric_dtype(self.data[c]) else
                self.data[c].str.len().max())
            for c in columns}

        var_labels = dict(zip(self.question_metadata.question_id,
                              self.question_metadata.full_text))

        value_labels = ([{question_id: {d['value_code']: d['full_label'] for d
                        in self.codebook[self.codebook.question_id ==
                                         question_id].to_dict(
                            orient='records')}} for question_id in
                        self.codebook.question_id.unique()])
        value_labels = {k: v for d in value_labels for k, v in d.items()}

        records = self.data

        with SavWriter(filepath, var_names, var_types,
                       value_labels, var_labels, ioUtf8=True,
                       *args, **kwargs) as writer:
            for record in records.values:
                writer.writerow(record)

    def check_schema_conversion_readiness(self):
        assert isinstance(self.question_metadata, pd.DataFrame), (
            "question_metadata should be a pandas DataFrame")
        assert "question_id" in self.question_metadata.columns, (
            "question_metadata should contain question_id column")
        assert "full_text" in self.question_metadata.columns, (
            "question_metadata should contain full_text column")
        assert "display_text" in self.question_metadata.columns, (
            "question_metadata should contain display_text column")
        assert "category" in self.question_metadata.columns, (
            "question_metadata should contain category column")
        assert "subcategory" in self.question_metadata.columns, (
            "question_metadata should contain subcategory column")
        assert "topic" in self.question_metadata.columns, (
            "question_metadata should contain topic column")
        assert "subtopic" in self.question_metadata.columns, (
            "question_metadata should contain subtopic column")
        assert "type_of_response" in self.question_metadata.columns, (
            "question_metadata should contain type_of_response column")
        assert "scale_id" in self.question_metadata.columns, (
            "question_metadata should contain scale_id column")
        assert "scale_name" in self.question_metadata.columns, (
            "question_metadata should contain scale_name column")
        assert "type_of_scale" in self.question_metadata.columns, (
            "question_metadata should contain type_of_scale column")

        assert isinstance(self.data, pd.DataFrame), (
            "data should be a pandas DataFrame")
        assert "id_in_study" in self.data.columns, (
            "data should contain id_in_study column")

        assert isinstance(self.codebook, pd.DataFrame), (
            "codebook should be a pandas DataFrame")
        assert "question_id" in self.codebook.columns, (
            "codebook should contain question_id column")
        assert "value_code" in self.codebook.columns, (
            "codebook should contain value_code column")
        assert "full_label" in self.codebook.columns, (
            "codebook should contain full_label column")
        assert "display_label" in self.codebook.columns, (
            "codebook should contain display_label column")

    def to_schema(self, study):
        studies = pd.DataFrame(
            columns=['study_id', 'study_name', 'start_date', 'end_date'],
            data=[[study.study_id, study.study_name,
                   study.start_date, study.end_date]])
        studies_questions = pd.DataFrame({
            'study_id': study.study_id,
            'question_id': self.data.columns.values
        })
        respondents = pd.DataFrame(
            {'id_in_survey': self.data.index.values,
             'respondent_id': self.data.index.values})
        studies_respondents = pd.DataFrame({
            'study_id': study.study_id,
            'respondent_id': self.data.index.values
        })
        responses = pd.melt(self.data.reset_index(), id_vars='index',
                            var_name='question_id')
        responses.rename(columns={'index': 'respondent_id'}, inplace=True)
        responses['question_value_id'] = (
            responses["question_id"] + "_" +
            responses["value"].astype(int).astype(str))
        responses = responses[['respondent_id', 'question_value_id']]
        responses['study_id'] = study.study_id

        schema = Schema(
            studies, studies_questions, respondents,
            studies_respondents, responses)

        return schema


class Schema(object):

    def __init__(self, studies, question_orders, questions, scales,
                 response_choices, closed_ended_responses,
                 open_ended_responses):

        assert isinstance(studies, pd.DataFrame), (
            "studies argument should be a pandas DataFrame")
        assert "study_id" in studies.columns, (
            "studies should contain study_id column")
        assert "study_name" in studies.columns, (
            "studies should contain study_name column")
        assert "start_date" in studies.columns, (
            "studies should contain start_date column")
        assert "end_date" in studies.columns, (
            "studies should contain end_date column")
        self.studies = studies

        assert isinstance(question_orders, pd.DataFrame), (
            "question_orders argument should be a pandas DataFrame")
        assert "study_id" in question_orders.columns, (
            "question_orders should contain study_id column")
        assert "question_id" in question_orders.columns, (
            "question_orders should contain question_id column")
        assert "order" in question_orders.columns, (
            "question_orders should contain order column")
        self.question_orders = question_orders

        assert isinstance(questions, pd.DataFrame), (
            "questions argument should be a pandas DataFrame")
        assert "question_id" in questions.columns, (
            "questions should contain question_id column")
        assert "full_text" in questions.columns, (
            "questions should contain full_text column")
        assert "display_text" in questions.columns, (
            "questions should contain display_text column")
        assert "category" in questions.columns, (
            "questions should contain category column")
        assert "subcategory" in questions.columns, (
            "questions should contain subcategory column")
        assert "topic" in questions.columns, (
            "questions should contain topic column")
        assert "subtopic" in questions.columns, (
            "questions should contain subtopic column")
        assert "type_of_response" in questions.columns, (
            "questions should contain type_of_response column")
        assert "scale_id" in questions.columns, (
            "questions should contain scale_id column")
        self.questions = questions

        assert isinstance(scales, pd.DataFrame), (
            "scales argument should be a pandas DataFrame")
        assert "scale_id" in scales.columns, (
            "scales should contain scale_id column")
        assert "scale_name" in scales.columns, (
            "scales should contain scale_name column")
        assert "type_of_scale" in scales.columns, (
            "scales should contain type_of_scale column")
        self.scales = scales

        assert isinstance(response_choices, pd.DataFrame), (
            "response_choices argument should be a pandas DataFrame")
        assert "response_choice_id" in response_choices.columns, (
            "response_choices should contain response_choice_id column")
        assert "question_id" in response_choices.columns, (
            "response_choices should contain question_id column")
        assert "scale_id" in response_choices.columns, (
            "response_choices should contain scale_id column")
        assert "full_label" in response_choices.columns, (
            "response_choices should contain full_label column")
        assert "display_label" in response_choices.columns, (
            "response_choices should contain display_label column")
        assert "value" in response_choices.columns, (
            "response_choices should contain value column")
        self.response_choices = response_choices

        assert isinstance(closed_ended_responses, pd.DataFrame), (
            "closed_ended_responses argument should be a pandas DataFrame")
        assert "study_id" in closed_ended_responses.columns, (
            "closed_ended_responses should contain study_id column")
        assert "question_id" in closed_ended_responses.columns, (
            "closed_ended_responses should contain question_id column")
        assert "respondent_id" in closed_ended_responses.columns, (
            "closed_ended_responses should contain respondent_id column")
        assert "value" in closed_ended_responses.columns, (
            "closed_ended_responses should contain value column")
        self.closed_ended_responses = closed_ended_responses

        assert isinstance(open_ended_responses, pd.DataFrame), (
            "open_ended_responses argument should be a pandas DataFrame")
        assert "study_id" in open_ended_responses.columns, (
            "open_ended_responses should contain study_id column")
        assert "question_id" in open_ended_responses.columns, (
            "open_ended_responses should contain question_id column")
        assert "respondent_id" in open_ended_responses.columns, (
            "open_ended_responses should contain respondent_id column")
        assert "text_response" in open_ended_responses.columns, (
            "open_ended_responses should contain text_response column")
        self.open_ended_responses = open_ended_responses

    def to_csv(self, studies_filepath,
               question_orders_filepath,
               questions_filepath,
               scales_filepath,
               response_choices_filepath,
               closed_ended_responses_filepath,
               open_ended_responses_filepath,
               encoding='utf-8', *args, **kwargs):
        self.studies.to_csv(
            studies_filepath, index=False, encoding=encoding,
            *args, **kwargs)
        self.question_orders.to_csv(
            question_orders_filepath, index=False, encoding=encoding,
            *args, **kwargs)
        self.questions.to_csv(
            questions_filepath, index=False, encoding=encoding,
            *args, **kwargs)
        self.scales.to_csv(
            scales_filepath, index=False, encoding=encoding,
            *args, **kwargs)
        self.response_choices.to_csv(
            response_choices_filepath, index=False, encoding=encoding,
            *args, **kwargs)
        self.closed_ended_responses.to_csv(
            closed_ended_responses_filepath, index=False, encoding=encoding,
            *args, **kwargs)
        self.open_ended_responses.to_csv(
            open_ended_responses_filepath, index=False, encoding=encoding,
            *args, **kwargs)

    def to_excel(self, studies_filepath,
                 question_orders_filepath,
                 questions_filepath,
                 scales_filepath,
                 response_choices_filepath,
                 closed_ended_responses_filepath,
                 open_ended_responses_filepath,
                 encoding='utf-8', *args, **kwargs):
        self.studies.to_excel(
            studies_filepath, index=False, encoding=encoding,
            *args, **kwargs)
        self.question_orders.to_excel(
            question_orders_filepath, index=False, encoding=encoding,
            *args, **kwargs)
        self.questions.to_excel(
            questions_filepath, index=False, encoding=encoding,
            *args, **kwargs)
        self.scales.to_excel(
            scales_filepath, index=False, encoding=encoding,
            *args, **kwargs)
        self.response_choices.to_excel(
            response_choices_filepath, index=False, encoding=encoding,
            *args, **kwargs)
        self.closed_ended_responses.to_excel(
            closed_ended_responses_filepath, index=False, encoding=encoding,
            *args, **kwargs)
        self.open_ended_responses.to_excel(
            open_ended_responses_filepath, index=False, encoding=encoding,
            *args, **kwargs)

    def to_sql(self, connection, *args, **kwargs):
        self.studies.to_sql(
            'studies', connection, *args, **kwargs)
        self.question_orders.to_sql(
            'question_orders', connection, *args, **kwargs)
        self.questions.to_sql(
            'questions', connection, *args, **kwargs)
        self.scales.to_sql(
            'scales', connection, *args, **kwargs)
        self.response_choices.to_sql(
            'response_choices', connection, *args, **kwargs)
        self.closed_ended_responses.to_sql(
            'closed_ended_responses', connection, *args, **kwargs)
        self.open_ended_responses.to_sql(
            'open_ended_responses', connection, *args, **kwargs)



from sqlalchemy import Column, Date, Float, ForeignKey, String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

Base = declarative_base()


class Study(Base):
    __tablename__ = 'studies'

    study_id = Column(String(10), primary_key=True)
    study_name = Column(String(255), nullable=False)
    start_date = Column(Date)
    end_date = Column(Date)

    def __init__(self, study_id, study_name, start_date, end_date):
        self.study_id = study_id
        self.study_name = study_name
        self.start_date = start_date
        self.end_date = end_date

    def __repr__(self):
        return "<Study(study_id = '%s', study_name='%s', start_date='%s', end_date='%s')>" % (
            self.study_id, self.study_name, self.start_date, self.end_date)


class QuestionOrder(Base):
    __tablename__ = 'question_orders'

    study_id = Column(
        String(10), ForeignKey('studies.study_id'), primary_key=True)
    question_id = Column(
        String(50), ForeignKey('questions.question_id'), primary_key=True)
    order = Column(Float)

    question = relationship("Question")
    study = relationship("Study")

    def __init__(self, study_id, question_id, order):
        self.study_id = study_id
        self.question_id = question_id
        self.order = order


class Scale(Base):
    __tablename__ = 'scales'

    scale_id = Column(String(50), primary_key=True)
    scale_name = Column(String(50))
    type_of_scale = Column(String(50))

    def __init__(self, scale_id, scale_name, type_of_scale):
        self.scale_id = scale_id
        self.scale_name = scale_name
        self.type_of_scale = type_of_scale


class Question(Base):
    __tablename__ = 'questions'

    question_id = Column(String(50), primary_key=True)
    full_text = Column(String)
    display_text = Column(String(50))
    category = Column(String(50))
    subcategory = Column(String(50))
    topic = Column(String(50))
    subtopic = Column(String(50))
    type_of_response = Column(String(50))
    scale_id = Column(String(10), ForeignKey('scales.scale_id'))

    scale = relationship("Scale")

    def __init__(self, question_id, full_text, display_text,
                 category, subcategory, topic, subtopic,
                 type_of_response, scale_id):
        self.question_id = question_id
        self.full_text = full_text
        self.display_text = display_text
        self.category = category
        self.subcategory = subcategory
        self.topic = topic
        self.subtopic = subtopic
        self.type_of_response = type_of_response
        self.scale_id = scale_id


class ResponseChoice(Base):
    __tablename__ = 'response_choices'

    response_choice_id = Column(String(50), primary_key=True)
    question_id = Column(String(50), ForeignKey('questions.question_id'))
    scale_id = Column(String(10), ForeignKey('scales.scale_id'))
    full_label = Column(String(50))
    display_label = Column(String(50))
    value = Column(Float)

    scale = relationship("Scale")
    question = relationship("Question")

    def __init__(self, response_choice_id, question_id, scale_id,
                 full_label, display_label, value):
        self.response_choice_id = response_choice_id
        self.question_id = question_id
        self.scale_id = scale_id
        self.full_label = full_label
        self.display_label = display_label
        self.value = value


class ClosedEndedResponse(Base):
    __tablename__ = 'closed_ended_responses'

    study_id = Column(
        String(10), ForeignKey('studies.study_id'), primary_key=True)
    respondent_id = Column(
        String(50), primary_key=True)
    question_id = Column(
        String(50), ForeignKey('questions.question_id'), primary_key=True)
    value = Column(Float)

    question = relationship("Question")
    study = relationship("Study")

    def __init__(self, study_id, respondent_id, question_id,
                 value):
        self.study_id = study_id
        self.respondent_id = respondent_id
        self.question_id = question_id
        self.value = value


class OpenEndedResponse(Base):
    __tablename__ = 'open_ended_responses'

    study_id = Column(
        String(10), ForeignKey('studies.study_id'), primary_key=True)
    respondent_id = Column(
        String(50), primary_key=True)
    question_id = Column(
        String(50), ForeignKey('questions.question_id'), primary_key=True)
    text_response = Column(String)

    question = relationship("Question")
    study = relationship("Study")

    def __init__(self, study_id, respondent_id, question_id,
                 text_response):
        self.study_id = study_id
        self.respondent_id = respondent_id
        self.question_id = question_id
        self.text_response = text_response



from setuptools import setup

setup(name='polarbear',
      version='0.0.0.dev6',
      description='logic for social science survey data',
      packages=['polarbear'],
      install_requires=[
            'pandas',
            'savReaderWriter',
            'sqlalchemy',
            'xlrd',
            'openpyxl'
        ]
      )



import pandas as pd
from polarbear import entities, models


def convert_dataset_to_schema(ds, mapping, study):
    questions_info = mapping.info.drop(
        ['type_of_scale', 'scale_exists_in_database',
         'scale_name'], axis=1)
    scales_info = mapping.info[
        ['type_of_scale', 'scale_exists_in_database',
         'scale_id', 'scale_name']]

    questions_info = questions_info[questions_info['exists_in_database'] == 0]
    scales_info = scales_info[scales_info['scale_exists_in_database'] == 0]

    studies = pd.DataFrame(
        columns=['study_id', 'start_date', 'end_date', 'study_name'],
        data=[[study.study_id, study.start_date,
               study.end_date, study.study_name]])

    question_orders = pd.DataFrame({
        'study_id': study.study_id,
        'question_id': [dict(zip(mapping.info.question_id,
                        mapping.info.database_question_id)).get(
                        item, item) for item in list(ds.data.columns)],
        'order': [float(i) for i in (range(1, len(ds.data.columns) + 1))]
    })

    questions = pd.DataFrame({
        'question_id': questions_info.database_question_id,
        'full_text': questions_info.full_text,
        'display_text': questions_info.display_text,
        'category': questions_info.category,
        'subcategory': questions_info.subcategory,
        'topic': questions_info.topic,
        'subtopic': questions_info.subtopic,
        'type_of_response': questions_info.type_of_response,
        'scale_id': questions_info.scale_id
    })

    scales = pd.DataFrame({
        'scale_id': scales_info.scale_id,
        'scale_name': scales_info.scale_name,
        'type_of_scale': scales_info.type_of_scale
    })

    response_choices = pd.DataFrame({
        'question_id': ds.codebook.question_id.replace(
            dict(zip(mapping.info.question_id,
                     mapping.info.database_question_id))),
        'scale_id': ds.codebook.question_id.replace(
            dict(zip(mapping.info.question_id, mapping.info.scale_id))),
        'full_label': ds.codebook.full_label,
        'display_label': ds.codebook.display_label,
        'value': [float(i) for i in ds.codebook.value_code]
    })

    response_choices = response_choices[response_choices['question_id'].isin(
        questions_info.database_question_id)].copy().reset_index(drop=True)
    response_choices['response_choice_id'] = \
        response_choices.question_id.str.cat(
            response_choices.value.astype(int).astype(str), sep='_')
    columns = models.ResponseChoice.__table__.columns.keys()
    response_choices = response_choices[columns]

    closed_ended_mask = mapping.info.type_of_response.isin(
        ['Scale', 'Multiple response'])
    closed_ended_questions = list(mapping.info[closed_ended_mask].question_id)
    open_ended_questions = list(mapping.info[~closed_ended_mask].question_id)

    ds.data['id_in_study'] = ds.data['id_in_study'].astype(int).astype(str)
    ds.data['respondent_id'] = study.study_id + '_' + ds.data['id_in_study']

    closed_ended_data = ds.data[['respondent_id'] + closed_ended_questions]
    closed_ended_responses = pd.melt(
        closed_ended_data, id_vars='respondent_id', var_name='question_id')
    closed_ended_responses['question_id'].replace(
        dict(zip(mapping.info.question_id,
                 mapping.info.database_question_id)),
        inplace=True)
    closed_ended_responses.value = closed_ended_responses.value.astype(float)
    closed_ended_responses['study_id'] = study.study_id
    columns = models.ClosedEndedResponse.__table__.columns.keys()
    closed_ended_responses = closed_ended_responses[columns]

    open_ended_data = ds.data[['respondent_id'] + open_ended_questions]
    open_ended_responses = pd.melt(
        open_ended_data, id_vars='respondent_id', var_name='question_id',
        value_name='text_response')
    open_ended_responses['question_id'].replace(
        dict(zip(mapping.info.question_id,
                 mapping.info.database_question_id)),
        inplace=True)
    open_ended_responses['study_id'] = study.study_id
    open_ended_responses['text_response'] = open_ended_responses['text_response'].astype(str)
    columns = models.OpenEndedResponse.__table__.columns.keys()
    open_ended_responses = open_ended_responses[columns]

    schemed = entities.Schema(
        studies, question_orders, questions, scales,
        response_choices, closed_ended_responses, open_ended_responses)

    return schemed
